package linked_list;

class Item<T> {

    private T value;
    private Item<T> next;

    public Item(T value, Item<T> next) {
        this.value = value;
        this.next = next;
    }

    public Item<T> next() {
        return next;
    }

    public void next(Item<T> next) {
        this.next = next;
    }

    public T value() {
        return value;
    }

    public void value(T value) {
        this.value = value;
    }
}


class LinkedList<T> {

    private Item<T> head;

    public LinkedList() {
        head = null;
    }

    // returns 0 on success, -1 on failure
    public int insert(T new_value) {
        head = new Item<T>(new_value, head);
        if (head != null) {
            return 0;
        }
        return -1;
    }

    // returns 0 on success, -1 on failure
    public int remove (T value_to_remove) {
        Item<T> marker = head;
        Item<T> temp = null;  // temp points to one behind as we iterate

        while (marker != null) {
            if (marker.value() == value_to_remove) {
                if (temp == null) { // marker is the first element in the list
                    if (marker.next() == null) {
                        head = null;
                        marker = null;
                    } else {
                        // FIXED: no need to create new item, just let head = marker.next()
                        head = marker.next();
                        marker = null;
                    }
                    return 0;
                } else {
                    temp.next (marker.next());
                    temp = null;
                    return 0;
                }
            }
            // FIXED: immediate segfault, not needed at all
            temp = marker;
            marker = marker.next();
        }

        return -1; // failure
    }

    public Item<T> find(T value) {
        // Fixed: will crash on empty list, skips first item
        Item<T> marker = head;
        while (marker != null) {
            if (marker.value() == value)
                return marker;
            // FIXED: advance the marker
            marker = marker.next();
        }
        return null;
    }

    public void print() {
        Item<T> marker = head;
        System.out.println("Current state of list:");
        // FIXED: Fails when list is empty
        while (marker != null) {
            System.out.println(marker.value());
            marker = marker.next();
        }
    }

    public void clear() {
        Item<T> marker = head;
        while (marker != null) {
            Item<T> temp = marker;
            // FIXED: doesn't actually clear the list
            marker = null;
            // FIXED: should be marker = temp.next()
            marker = temp.next();
        }
    }

    public void run() {
        // TODO: Should be test cases

        LinkedList<Integer> list = new LinkedList<Integer>();

        System.out.println("Adding elements:");
        list.insert(1);
        list.insert(2);
        list.insert(3);
        list.insert(4);

        Item<Integer> query = list.find(3);
        System.out.println("Searching 3: found " + query.value().toString() );

        list.print();

        System.out.println("Removing elements:");
        list.remove(2);
        list.print();

        list.remove(4);
        list.print();

        list.remove(1);
        list.print();

        list.remove(2);
        list.print();

        list.remove(3);
        list.print();

        list.clear();
    }

    public static void main(String args[]) {
        LinkedList debugMe = new LinkedList();
        debugMe.run();
    }
}
